/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"
#include "udb.h"
#include "file.h"
#include "tag.h"
#include "str.h"
#include "vstr.h"
#include "last_log.h"
#include "partdir.h"
#include "btpasswd.h"
#include "group.h"
#include "partfile.h"
#include "dumputil.h"
#include "initcmd.h"
#include "xmlparse.h"

#include <string.h>
#include <ctype.h>

#ifdef HAVE_DIRENT_H
#include <dirent.h>
typedef struct dirent direntry;
#else
typedef struct direct direntry;
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

uid_t max_uid;		/* Max uid restored in this dump */
int verbose= FALSE;

/* FINDCONF - Find the partlist entry corresponding to a give conference name.
 */

struct plist *findconf(char *conf)
{
    struct plist *pl;
    int conflen= strlen(conf);

    for (pl= partlist; pl != NULL; pl= pl->next)
	if (pl->cflen <= conflen && !strncmp(pl->cf, conf, pl->cflen))
	    return pl;
    return NULL;
}


/* FIND_GID:  Given a group name, return the group id.
 */

gid_t find_gid(char *name)
{
    int i;

    for (i= 0; i < maxgrp; i++)
	if (grplist[i] != NULL && !strcmp(name,grplist[i]))
	    return i;

    die("Could not find group %s", name);
}


void writefile(char *login, char *directory, char *filename, char *content)
{
    char path[BFSZ];
    FILE *fp;

#ifdef PART_DIR
    if (strcmp(filename,".cflist") != 0 && strcmp(filename,".backtalk") != 0)
	bbspartpath(path,login,filename);
    else
#endif
	homepartpath(path,directory,filename);
    
    if (verbose)
	fprintf(stderr,"Writing File %s:\n%s\n", path,
		content != NULL ? content:"(empty)");

    if ((fp= fopen(path, "w")) == NULL)
    {
	fprintf(stderr, "Could not open %s to write\n", path);
	return;
    }

    if (content != NULL)
	while (*content != '\0')
	    fputc(*(content++), fp);

    fclose(fp);
}


/* GET_NOISE and NEXT_UID():  We don't want to generate new validation codes
 * or uid number, so we replace the real get_noise() and next_uid() with
 * versions that return the values we read from the file.
 */

char *evalid;
char *get_noise() {return evalid;} /* SUBSTITUTE */


/* MAKE_USER: Given a login ID and some tagfile data, create a user.  Returns
 * 1 if the user already existed.  The tagdata is freed by this.
 */

int make_user(char *login, char *tagdata, char *directory)
{
    rtag *rt= NULL;
    int rc, status= 0, privacy= 1, eprivacy= 0;
    char *var, *val;
    char *usrname= NULL, *grps= NULL;
    char *encpasswd= NULL, *email= NULL;
    char *passwd= NULL;
    time_t laston;
    gid_t gid;
    int have_uid= FALSE, have_gid= FALSE;
    struct btpasswd *pw;
    uid_t uid;
    VSTRING dotbacktalk;

    evalid= NULL;
    vs_new(&dotbacktalk, BFSZ);

    if (login == NULL)
	die("No login");

    if ((pw= getdbnam(login)) != NULL)
    {
	fprintf(stderr,"user %s already exists.  Skipping\n", login);
	free(tagdata);
	return 1;
    }

    /* Fake opening a tag file */
    if (tagdata != NULL)
    {
	rt= (rtag *) malloc(sizeof(rtag));
	rt->buf= tagdata;
	rt->p= rt->buf;
    }

    if (verbose) fprintf(stderr,"LOGIN=%s\n",login);
    while ((rc= get_tag(rt,&var,&val)) > 0)
    {
	if (rc > 2) rc-= 2;
	if (verbose) fprintf(stderr,"%s => %s\n", var, val);
	
	if (!strcmp(var,"usrname"))
	    usrname= val;
	else if (!strcmp(var,"dirpath"))
	    /*ignore it*/;
	else if (!strcmp(var,"grps"))
	    grps= val;
	else if (!strcmp(var,"encpasswd"))
	    encpasswd= val;
	else if (!strcmp(var,"passwd"))
	    passwd= val;
	else if (!strcmp(var,"email"))
	    email= val;
	else if (!strcmp(var,"evalid"))
	    evalid= val;
	else if (!strcmp(var,"usrid"))
	{
	    uid= atoi(val);
	    if (uid > max_uid) max_uid= uid;
	    have_uid= TRUE;

#if defined(ID_GETUID) && !defined(UID_MAP_FILE)
	    /* If we have a way to look up users by uid, the use that to
	     * check that uid is not duplicated.  We usually don't have
	     * this capability though. */
	    if ((pw= getdbuid(uid)) != NULL)
	    {
		fprintf(stderr,
		    "user %s has uid %d already in use by %s.  Skipping.\n",
		    login, uid, pw->name);
		goto done;
	    }
#endif
	}
	else if (!strcmp(var,"usrstat"))
	    status= atoi(val);
	else if (!strcmp(var,"privacy"))
	    privacy= atoi(val);
	else if (!strcmp(var,"eprivacy"))
	    eprivacy= atoi(val);
	else if (!strcmp(var,"grpname"))
	{
	    gid= find_gid(val);
	    have_gid= TRUE;
	}
	else
	{
	    /* Add unknown token to list of things to put in .backtalk */
	    vs_cat(&dotbacktalk, var);
	    *vs_inc(&dotbacktalk)= '=';
	    if (rc == 1)
	    {
		char *p;
		/* Write string tag value to memory */
		*vs_inc(&dotbacktalk)= '(';
		for (p= val; *p != '\0'; p++)
		    switch (*p)
		    {
		    case '\r':                                   break;
		    case '\n': vs_cat_n(&dotbacktalk,"\\n",2);  break;
		    case '\\': vs_cat_n(&dotbacktalk,"\\\\",2); break;
		    case '(':  vs_cat_n(&dotbacktalk,"\\(",2);  break;
		    case ')':  vs_cat_n(&dotbacktalk,"\\)",2);  break;
		    default:   *vs_inc(&dotbacktalk)= *p;	 break;
		    }
		*vs_inc(&dotbacktalk)= ')';
	    }
	    else
	    {
		/* Write integer tag value to memory */
		vs_cat(&dotbacktalk, val);
	    }
	    *vs_inc(&dotbacktalk)= '\n';
	}
    }

    if (!have_uid)
	die ("No uid for %s",login);
    if (!have_gid)
	die ("No primary gid for %s",login);
    if (encpasswd == NULL && passwd == NULL)
	encpasswd= "x";

    if (verbose)
	fprintf(stderr,"Creating %s\n uid=%d\n gid=%d\n passwd=%s\n name=%s\n"
		       " status=%d\n dir=%s\n email=%s\n evalid=%s\n"
		       " privacy=%d\n eprivacy=%d\n",
			login,uid,gid,passwd?passwd:encpasswd,
			usrname?usrname:"NULL",status,directory,email,
			evalid?evalid:"NULL", privacy, eprivacy);

    /* Note that the validation code is passed into this in a roundabout way */
    adduser(login,
	   gid,
	   passwd ? passwd : "\r\n",
	   usrname ? usrname : login,
	   status,
	   directory,
	   email,
	   evalid == NULL,
	   privacy,
	   eprivacy,
	   TRUE, uid);

    /* If we had an encrypted password, we need to save it now.  Clunky,
     * but that's life */
    if (passwd == NULL)
    {
	char pass[BFSZ], *np= pass;

	if ((pw= getdbnam(login)) == NULL)
	    die("Can't found account %s after creating it",login);

	if (status > 0)
	    *(np++)= (status == 1) ? '?' : '*';
	strcpy(np, encpasswd);
	change_passwd(pw, pass);
    }

    /* Write out the .backtalk file */
    if (vs_size(&dotbacktalk) > 0)
    {
	*vs_inc(&dotbacktalk)= '\0';
	writefile(login,directory,".backtalk",dotbacktalk.begin);
    }
    vs_destroy(&dotbacktalk);

done:
    close_tag(rt);
    return 0;
}


/* PART_DATA:  Given a participation data structure, and some tag data,
 * add alias and other top-level info to pf for conf.  The tagdata is
 * freed by this.
 */

void part_data(struct pfc *part, char *tagdata)
{
    rtag *rt= NULL;
    int rc;
    char *var, *val;

    if (tagdata == NULL)
	return;
    if (part == NULL)
    {
	free(tagdata);
	return;
    }

    /* Fake opening a tag file */
    rt= (rtag *) malloc(sizeof(rtag));
    rt->buf= tagdata;
    rt->p= rt->buf;

    while ((rc= get_tag(rt,&var,&val)) > 0)
    {
	if (!strcmp(var,"alias"))
	    part->alias= strdup(val);
	else if (rc == 1 || rc == 3)
	    etc_str(&part->conf_etc, var, val, TRUE);
	else
	    etc_int(&part->conf_etc, var, atoi(val));
    }

    close_tag(rt);
}


/* ITEM_DATA:  Given a participation data structure, an item number, and some
 * tag data, add information about that item to the participation data
 * structure.  The tagdata is freed by this.
 */

void item_data(struct pfc *part, int item, char *tagdata)
{
    rtag *rt= NULL;
    struct pfe *pitem;
    int rc;
    char *var, *val;

    if (tagdata == NULL)
	return;
    if (part == NULL || item < 1)
    {
	free(tagdata);
	return;
    }

    /* Create the pf data structure entry for this item */
    if (find_pf(part, item, TRUE, &pitem) == 0)
    {
	fprintf(stderr,"Participation information for %s conference "
		"item %d already exists.  Not modifying.\n",
		part->conf, item);
	free(tagdata);
	return;
    }

    /* Fake opening a tag file */
    rt= (rtag *) malloc(sizeof(rtag));
    rt->buf= tagdata;
    rt->p= rt->buf;

    while ((rc= get_tag(rt,&var,&val)) > 0)
    {
	if (!strcmp(var,"lastresp"))
	    pitem->resp_no= atoi(val);
	if (!strcmp(var,"lastdate"))
	    pitem->date= atoi(val);
	if (!strcmp(var,"forgotten"))
	    pitem->forgotten= atoi(val);
	if (!strcmp(var,"favorite"))
	    pitem->favorite= atoi(val);
	if (!strcmp(var,"title"))
	    pitem->my_title= strdup(val);
	else if (rc == 1 || rc == 3)
	    etc_str(&pitem->etc, var, val, TRUE);
	else
	    etc_int(&pitem->etc, var, atoi(val));
    }

    close_tag(rt);
}


#define CN_NONE 0
#define CN_USER 1
#define CN_PART 2
#define CN_ITEM 3

int main(int argc, char **argv)
{
    struct btpasswd *pw;
    int load_part= TRUE;
    int i,j,type;
    char *data, *var, *val;
    struct usl {
	char *u;
	struct usl *next;
    } *u, *last= NULL, *userlist= NULL;
    char *version;
    char *login= NULL;
    char *conf= NULL;
    struct pfc *part= NULL;
    char *content= NULL;
    char *filename= NULL;
    int item= -1;
    int context= CN_NONE;
    char directory[BFSZ];

    max_uid= 0;

    init_cmd(argv[0],1);

    for (i= 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	    for (j= 1; argv[i][j] != '\0'; j++)
		switch (argv[i][j])
		{
		    case 'p':  load_part= FALSE; break;
		    case 'v':  verbose= TRUE; break;
		    default: goto usage;
		}
	else
	{
	    /* Insert user into user name list */
	    u= (struct usl *)malloc(sizeof(struct usl));
	    u->u= argv[i];
	    u->next= NULL;
	    if (last == NULL)
		userlist= u;
	    else
		last->next= u;
	    last= u;
	}
    }

    if (load_part)
	/* Build a list of participation file names that go with each
	 * conference name. */
	mkpartlist();

    /* Initialize the group list structure */
    for_groups(add_group);

    /* Read the file */
    while ((type= getinput(stdin,&data)) != END_OF_DATA)
    {
	switch (type)
	{
	case START_ELEMENT:
	    /* For start tags we mostly just stash away some attribute values */
	    if (verbose) fprintf(stderr,"START %s\n",data);

	    if (!strcmp(data,"USERDATA"))
	    {
		/* Save "version" attribute, ignore all others */
		while (getattr(stdin,&var,&val))
		{
		    if (!strcmp(var,"version"))
			version= val;
		    else
			free(val);
		    free(var);
		}
	    }
	    else if (!strcmp(data,"USER"))
	    {
		/* Save "login" attribute, ignore all others */
		context= CN_USER;
		if (login != NULL) free(login);
		login= NULL;
		while (getattr(stdin,&var,&val))
		{
		    if (!strcmp(var,"login"))
			login= val;
		    else
			free(val);
		    free(var);
		}
		if (login == NULL)
		    die("No login name given for user");
	    }
	    else if (!strcmp(data,"PARTICIPATION"))
	    {
		/* Save "conf" attribute, ignore all others */
		context= CN_PART;
		if (conf != NULL) free(conf);
		conf= NULL;
		while (getattr(stdin,&var,&val))
		{
		    if (!strcmp(var,"conf"))
			conf= val;
		    else
			free(val);
		    free(var);
		}
		if (conf == NULL)
		    die("No conf name given for participation data");

		if (load_part && login)
		{
		    /* Look up conference in our list */
		    struct plist *pl= findconf(conf);
		    if (pl == NULL)
		    {
			fprintf(stderr,"Not creating participation file for "
				"%s conference for user %s -- "
				"Conference does not exist\n",
				conf, login);
			part= NULL;
		    }
		    else
		    {
			/* Create the participation file */
			part= (struct pfc *)malloc(sizeof(struct pfc));
			part->snooping= FALSE;
#ifdef PART_DIR
			bbspartpath(part->part_name,login,pl->pf);
#else
			homepartpath(part->part_name,directory,pl->pf);
#endif
			part->user= login;
			part->conf= conf;
			/* Don't set part->confdir - we don't want to
			 * initialize the participation information based
			 * on conference new_isel/new_since settings.
			 */
			part->confdir= NULL;
			if (!loadpart(part))
			{
			    fprintf(stderr,"Not creating participation file "
				    "for %s conference for user %s -- "
				    "Participation file already exists\n",
				    conf, login);
			    freepart(part);
			    closepf(part);
			    free(part);
			    part= NULL;
			}
			else
			    part->user= part->conf= NULL;
		    }
		}
	    }
	    else if (!strcmp(data,"ITEM"))
	    {
		context= CN_ITEM;
		item= -1;
		/* Save item number attribute, ignore all others */
		while (getattr(stdin,&var,&val))
		{
		    if (!strcmp(var,"number") && (item= atoi(val)) <= 0)
			die("Bad Item Number");
		    free(val);
		    free(var);
		}
		if (item < 0)
		    die("No number given for item data");
	    }
	    else if (!strcmp(data,"FILE"))
	    {
		context= CN_USER;
		if (filename != NULL) free(filename);
		filename= NULL;
		/* Save file name attribute, ignore all others */
		while (getattr(stdin,&var,&val))
		{
		    if (!strcmp(var,"name"))
			filename= val;
		    else
			free(val);
		    free(var);
		}
		if (filename == NULL)
		    die("No name given for file");
	    }
	    else
	    {
		/* Some other tag - just discard all attributes */
		while (getattr(stdin,&var,&val))
		{
		    free(var);
		    free(val);
		}
	    }

	    if (data != NULL) free(data);

	    break;

	case END_ELEMENT:

	    /* End tags are where we usually do some processing */
	    if (data == NULL)
		die("File Format Error");
	    if (!strcmp(data, "DATA"))
	    {
		if (context == CN_USER)
		{
		    if (login != NULL)
		    {
			if (make_user(login,content,directory))
			{
			    /* If user already existed, don't load him */
			    free(login);
			    login= NULL;
			}
			content= NULL;
		    }
		}
		else if (context == CN_PART)
		{
		    /* Add alias and other top-level info to pf for conf */
		    if (part != NULL)
		    {
			part_data(part,content);
			content= NULL;
		    }
		}
	    }
	    else if (!strcmp(data, "ITEM"))
	    {
		/* add item 'item' data in 'context' to pf for 'conf' */
		if (part != NULL)
		{
		    item_data(part,item,content);
		    content= NULL;
		}
	    }
	    else if (!strcmp(data, "PARTICIPATION"))
	    {
		/* write out participation file */
		if (part != NULL)
		{
		    writepart(part);
		    closepf(part);
		    free(part);	/* Stuff pointed to has already be freed */
		    part= NULL;
		}
	    }
	    else if (!strcmp(data, "FILE"))
	    {
		/* write file */
		if (login != NULL)
		{
		    writefile(login,directory,filename,content);
		    free(content);
		    content= NULL;
		}
	    }
	    else
	    {
		/* Ignore other tags */
		if (content != NULL)
		{
		    free(content);
		    content= NULL;
		}
	    }

	    if (data != NULL)
		free(data);

	    break;

	case CONTENT:
	    /* Just store away any content */
	    if (verbose) fprintf(stderr,"CONTENT:\n%s\n",data);
	    if (content != NULL) free(content);
	    content= data;
	    break;
	}
    }

    /* Max sure next_uid is bigger than anything we loaded */
    if (next_uid(0) <= max_uid) next_uid_set(max_uid+1);

    exit(0);
usage:
    fprintf(stderr,"usage: %s [-pv] [<user>...]\n", argv[0]);
    exit(1);
}


/* Some dummy routines that never get called, but need to exist to let
 * things link.  Most of these are called by newpart() which will never
 * be called because we always open participation files in snoop mode.
 */

void seek_item(int x) {}   /* DUMMY */
int next_item(int *a,int *b,int *c,time_t *d,time_t *e) {return 1;}  /*DUMMY*/
int in_sel(int i, char *s) {return 0;} /* DUMMY */
char *user_fname() {return "";} /* DUMMY */
void wrttag(FILE *fp, char *name, Token *t) {}  /* DUMMY */
